
module Plugins.FOL.Gandalf (gandalfProof) where

import Data.Char
import Data.List
import Text.ParserCombinators.Parsec
import Text.PrettyPrint

import Utils.Monads

{--------------------------------------------------------------------------
    Pretty proof
 --------------------------------------------------------------------------}

gandalfProof :: String -> String
gandalfProof output = unlines $ concatMap disp $ proof
    where
	disp s = [show $ parseClause s]
	proof =
	    takeWhile ("END OF PROOF" /=)
	    $ drop 1
	    $ dropWhile ("START OF PROOF" /=)
	    $ lines output


{--------------------------------------------------------------------------
    Lexer
 --------------------------------------------------------------------------}

data Token  = TkInt Int
	    | TkJustification String
	    | TkId String
	    | TkSym Char
	    | TkContradiction
    deriving (Eq, Show)

isSym c = elem c "().,-|?"
isId c	= isAlphaNum c || elem c "_^'"

contradiction = "contradiction"

lexer :: String -> [Token]
lexer s
    | contradiction `isPrefixOf` s
		    = TkContradiction : lexer (drop (length contradiction) s)
lexer ('[':s)	    = case span (/=']') s of
			(s1,_:s2)   -> TkJustification s1 : lexer s2
lexer (c:s)
    | isSpace c = lexer s
    | isSym c	= TkSym c : lexer s
    | isDigit c = case span isDigit s of
			(s1,s2)	-> TkInt (read $ c:s1) : lexer s2
    | isId c	= case span isId s of
			(s1,s2)	-> TkId (c:s1) : lexer s2
lexer []	= []
lexer s		= error $ "lex error on input: " ++ s

{--------------------------------------------------------------------------
    Parser
 --------------------------------------------------------------------------}

type P = GenParser Token ()

data Clause	= Clause Int String [Literal]
data Literal	= Neg Atom
		| Pos Atom
		| Unknown
		| Contradiction
data Atom	= Predicate String [Term]
data Term	= Var String
		| Fun String [Term]

instance Show Clause where
    show = show . pretty

parseClause s =
    case parse clauseP "" (lexer s) of
	Right c	    -> c
	Left err    -> error $ show err

thisP t = tokenPrim show (\p _ _ -> p) f
    where
	f t'	| t' == t   = Just ()
	f _		    = Nothing

symP c = thisP (TkSym c)

clauseP :: P Clause
clauseP =
    do	TkInt n		    <- anyToken
	TkJustification j   <- anyToken
	ls <-	fmap (const [Contradiction]) (thisP TkContradiction)
		<|> fmap (const [Unknown]) (symP '?')
		<|> do	ls <- chainr1 (fmap (:[]) litP) orP
			symP '.'
			return ls
	return $ Clause n j ls
    where
	orP = const (++) <$> symP '|'

litP :: P Literal
litP =
    do	symP '-'
	Neg <$> atomP
    <|>	Pos <$> atomP

atomP :: P Atom
atomP =
    do	TkId x <- anyToken
	ts  <- argsP
	return $ Predicate x ts

argsP = option [] $
	    between (symP '(') (symP ')')
		    (fmap (const []) (thisP (TkInt 0))
			<|>
		     sepBy termP (symP ',')
		    )

termP :: P Term
termP =
    do	TkId x <- anyToken
	if isVar x
	    then return $ Var x
	    else do ts <- argsP
		    return $ Fun x ts
    where
	isVar (c:s) = isUpper c
	isVar _	    = False

{--------------------------------------------------------------------------
    App view
 --------------------------------------------------------------------------}

data AppTerm
	= AppVar String [Term]
	| AppFun String [Term]

appView (Var x) = AppVar x []
appView (Fun "capp_" [s,t]) = app (appView s) t
    where
	app (AppVar x ts) t = AppVar x (ts ++ [t])
	app (AppFun x ts) t = AppFun x (ts ++ [t])
appView (Fun f ts)	    = AppFun f ts

{--------------------------------------------------------------------------
    Pretty
 --------------------------------------------------------------------------}

class Pretty a where
    pretty	:: a -> Doc
    prettyPrec	:: Int -> a -> Doc

    pretty	    = prettyPrec 0
    prettyPrec _    = pretty

instance Pretty Int where
    pretty = text . show

instance Pretty Clause where
    pretty (Clause n j ls) =
	pretty n <+>
	    sep	[ fsep $ intersperse (text "\\/") $ map pretty ls
		, prJ j
		]
	where
	    prJ ""  = empty
	    prJ j   = text "[" <> text j <> text "]"

instance Pretty Literal where
    pretty (Neg (Predicate "equal" [s,t])) = prettyOp 2 (text "!=") s t 0
    pretty (Pos (Predicate "equal" [s,t])) = prettyOp 2 (text "==") s t 0
    pretty (Neg a)    = text "-" <> prettyPrec 9 a
    pretty (Pos a)    = pretty a
    pretty Unknown	    = text "?"
    pretty Contradiction    = text "_|_"

instance Pretty Atom where
    prettyPrec p (Predicate x [s,t])
	| op__ `isPrefixOf` x = prettyOp 6 (prettyPred x) s t p
    prettyPrec p (Predicate x ts) = pApp p (prettyPred x) ts

prettyOp p' op x y p =
    mparens (p>=p') $
	fsep
	    [ prettyPrec p' x
	    , nest 2 $ op
	    , prettyPrec p' y
	    ]

instance Pretty Term where
    prettyPrec p t =
	case appView t of
	    AppFun "cproj_" (t:pr:ts)	-> pApp p (pProj t pr) ts
	    AppFun x [s,t]
		| ('c':op__) `isPrefixOf` x
					-> prettyOp 6 (prettyFun x) s t p
	    AppFun x ts			-> pApp p (prettyFun x) ts
	    AppVar x ts			-> pApp p (prettyVar x) ts

mparens b = if b then parens else id

pProj t (Fun p []) = prettyPrec 10 t <> text s
    where
	s = case p of
		"cfst_"	-> ".1"
		"csnd_"	-> ".2"

pApp p d [] = d
pApp p d ts = mparens (p>9) $ d <+> argList ts
    where
	argList ts = fsep $ map (prettyPrec 10) ts

prettyVar   = text -- . ('#':)
prettyFun   = text . reop . up . tail
prettyPred  = text . reop . up

up ('^':c:s)	= toUpper c : up s
up (c:s)	= c : up s
up []		= []

op__ = "op__"

reop s
    | op__ `isPrefixOf` s   = parseOp $ drop (length op__) s
    | otherwise		    = s
    where

	trans _ ""		    = ""
	trans (op,name) s@(c:s')
	    | name `isPrefixOf` s   = op : trans (op,name) (drop (length name) s)
	    | otherwise		    = c : trans (op,name) s'

	parseOp s = foldr trans s ops

	ops =
	    [ ('!'  , "bang")
	    , ('#'  , "hash")
	    , ('$'  , "dollar")
	    , ('%'  , "percent")
	    , ('&'  , "amp")
	    , ('*'  , "times")
	    , ('+'  , "plus")
	    , ('.'  , "dot")
	    , ('/'  , "slash")
	    , ('<'  , "less")
	    , ('='  , "eq")
	    , ('>'  , "greater")
	    , ('?'  , "question")
	    , ('@'  , "at")
	    , ('\\' , "backslash")
	    , ('^'  , "hat")
	    , ('|'  , "bar")
	    , ('-'  , "minus")
	    , ('~'  , "twiddle")
	    , (':'  , "colon")
	    ]

