{-# OPTIONS -cpp #-}
{-# OPTIONS -fglasgow-exts -fallow-overlapping-instances #-}

module Plugins.QuickCheck where

import Prelude hiding	    ( print, putStr, putStrLn )
import IO.Output

import Control.Monad
import Control.Monad.Trans
import Control.Monad.Reader
import Control.Monad.State

#if __GLASGOW_HASKELL__ < 603
import qualified Debug.QuickCheck as Q
#else
import qualified Test.QuickCheck as Q
#endif

import Text.PrettyPrint
import System.Random
import System.IO hiding	    ( print, putStr, putStrLn )

import Syntax.Internal
import Syntax.Values
import Syntax.Pretty

import TypeChecker
import TypeChecker.Monad
import TypeChecker.Monad.Errors
import TypeChecker.Monad.Definitions
import TypeChecker.UnfoldConstants
import TypeChecker.Values
import TypeChecker.Haskell
import TypeChecker.Abstractions

import Utils.Error
import Utils.Names
import Utils.Monads
import Utils.Pairs
import Utils.Map

import Plugins

-- The plugin -------------------------------------------------------------

plugin :: SomePlugin
plugin = mkPlugin $
    Plugin  { plName		= "qc"
            , plHelp            = helpQuickCheck
	    , plInit		= initQC
	    , plFinal		= testGoals
	    , plCheck		= checkQC
	    , plInfer		= \_ _ _ _ -> failure noInferenceForQCError
	    , plOptions		= simpleOptions []
	    , plDefaultOptions	= ()
	    }

{-
-- Constants --------------------------------------------------------------

-- Random seeds
qc_StdGen	= "StdGen"
qc_split	= "split"
qc_pick		= "pick"
qc_seed		= "#seed#"

v_StdGen	= VDef qc_StdGen (DefVal 0 0 [] [] v_Set)
v_seed		= VDef qc_seed (DefVal 0 0 [] [] v_StdGen)

-- Generators
qc_Gen		= "Gen"
qc_monadGen	= "monadGen"
qc_return	= "return"
qc_bind		= "bind"
qc_oneof	= "oneof"
qc_sized	= "sized"
qc_frequency	= "frequency"
qc_variant	= "variant"
qc_promote	= "promote"
qc_elements	= "elements"

-- Properties
qc_Prop		= "Prop"
qc_Eq		= "Eq"
qc_True		= "True"
qc_False	= "False"
qc_collect	= "collect"
qc_label	= "label"
qc_classify	= "classify"
qc_trivial	= "trivial"
qc_forAll	= "forAll"

-- Useful datatypes
qc_nil		= "nil"
qc_cons		= ":"
qc_Nat		= "Nat"
qc_zero		= "zero"
qc_succ		= "succ"
qc_true		= "true"
qc_false	= "false"

-- Unfolding --------------------------------------------------------------

qcUnfoldConstants = genUnfoldConstants
			TopLevel
			(\x d vs -> x `elem` qcConstants
				    || isGenReturnOrBind x d vs)
			(const Nothing)
    where
	qcConstants =	[ qc_collect, qc_label, qc_classify, qc_trivial
			, qc_forAll, qc_oneof, qc_sized
			, qc_frequency, qc_variant, {-qc_promote,-} qc_elements
			]
{-	isGenReturnOrBind x (DefVal _ _ _) (IsApp g : _)
	    | x == qc_return || x == qc_bind	=
		case vAppView $ unfoldConstants g of
		    VAVDef c _ [_]  -> c == qc_Gen
		    _		    -> False
-}	isGenReturnOrBind x (DefVal _ _ _ _ _) (IsProj VFst : _) =
	    x == qc_monadGen
	isGenReturnOrBind x (DefVal _ _ _ _ _) (IsProj VSnd : _) =
	    x == qc_monadGen
	isGenReturnOrBind _ _ _ = False

-- Random -----------------------------------------------------------------

pick :: [Value] -> StdGen -> Value
pick xs r = xs !! fst (randomR (0, length xs - 1) r)

toStdGen r v =
    case vAppProjView $ unfoldConstants v of
	VAPVDef c _ [IsApp w, IsProj p]
	    | c == qc_split	-> pr p $ split $ toStdGen r w
	VAPVDef c _ []
	    | c == qc_seed	-> r
	_			-> error $ "Not a canonical StdGen: " ++ show v
    where
	pr VFst = fst
	pr VSnd = snd

-}
-- Properties -------------------------------------------------------------

data QCProp = QCProp Value Q.Property
-- Init -------------------------------------------------------------------

initQC _ = return []

-- Final ------------------------------------------------------------------

testGoals _ [] = return ()
testGoals _ props = liftIO $
    do	putStrLn "=== Testing QuickCheck properties ===================="
	hSetBuffering stdout NoBuffering
	mapM_ testProp props
    where
	testProp (QCProp v prop) =
	    do	print $ sep [ text "Testing property"
			    , nest 2 $ pretty v
			    ]
		Q.quickCheck prop
{-
-- Checking ---------------------------------------------------------------

hasTypeProp v =
    case unfoldConstants $ valueType v of
	VSort (VSUser prop _)	-> prop == qc_Prop
	_			-> False
-}
checkQC _ s _ ts g = return (vTop, s)
{-
    | hasTypeProp g	=
	do  d <- lookupDef qc_Gen
	    gens <- mkGens (VDef qc_Gen d) ts g
	    let p = buildProp gens g
	    return (VTop g, s ++ [QCProp g p])
    | otherwise	= failure $ notAPropError g
    where
	buildProp :: [Q.Gen Value] -> Value -> Q.Property
	buildProp [] goal = toHaskell goal
	buildProp (g:gs) goal =
	    case qcUnfoldConstants goal of
		VPi _ f _   ->
		    Q.forAll g $ \v -> buildProp gs (f `clApp` v)
		_	    -> error $ "Internal error: Checked QuickCheck property was bad: " ++ show goal

	mkGens d [] g = return []
	mkGens d (t:ts) g =
	    case unfoldConstants g of
		VPi w f _   ->
		    do	gen <- checkType t (d `vApp` w)
			x <- fresh (clName f)
			(toHaskell gen :) <$> mkGens d ts (f `clApp` VId x w)
		_	    -> failure $ toManyArgumentsError g

-- * Decidable equality on values

-- | Does not work for things with closures. But we actually have
--   boolean equality on values, even for closures so we could in
--   principle use it. What would it mean to test if two functions
--   are equal?
equal :: Value -> Value -> Bool
v `equal` w = unfoldConstants v === unfoldConstants w
    where
	VDef x _	    === VDef y _	    = x == y
	VId x _		    === VId y _		    = x == y    -- We can compare variables...
	VApp v v'	    === VApp w w'	    = equal v w && equal v' w'
	VSort s1	    === VSort s2	    = s1 == s2
	VTop _		    === VTop _		    = True
	VPair v v' _	    === VPair w w' _	    = equal v w && equal v' w'
	VSigma _ _ _ _ _    === _		    = error "No decidable equality for Sigma"
	_		    === VSigma _ _ _ _ _    = error "No decidable equality for Sigma"
	VPi _ _ _	    === _		    = error "No decidable equality for Pi"
	_		    === VPi _ _ _	    = error "No decidable equality for Pi"
	VLam _		    === _		    = error "No decidable equality for Lambda"
	_		    === VLam _		    = error "No decidable equality for Lambda"
	_		    === _		    = False

instance Extractable Q.Property where
    toHaskell goal =
	case vAppView $ qcUnfoldConstants goal of
	    VAVDef c _ [_, v, w]
		| c == qc_Eq	    -> Q.property $ v `equal` w
	    VAVDef c _ [_, g, f]
		| c == qc_forAll    -> Q.forAll (toGen g)
					    $ \v -> toProp (f `vApp` v)
	    VAVDef c _ []
		| c == qc_True	    -> Q.property True
		| c == qc_False	    -> Q.property False
	    VAVDef c _ [_,v,p]
		| c == qc_collect   -> Q.collect    (unfoldAllConstants v)
						    (toProp p)
	    VAVDef c _ [_,l,p]
		| c == qc_label	    -> Q.label (show l) (toProp p)
	    VAVDef c _ [_,b,l,p]
		| c == qc_classify  -> Q.classify   (toHaskell b) (show l)
						    (toProp p)
	    VAVDef c _ [b,p]
		| c == qc_trivial   -> Q.trivial (toHaskell b) (toProp p)
	    VAVSigma _ w _ f _
		| clName f == "_"   -> Q.property (mkBoolProp w &&
						    mkBoolProp (f `clApp` VId "_" w)
						)
	    VAVPi w f _
		| clName f == "_"
				-> mkBoolProp w Q.==>
				    toProp (f `clApp` VId "_" w)
				    -- Note: I wonder if this is the right
				    -- choice? Perhaps it is, if we arrive
				    -- at something unrecognizable it is a
				    -- false property. At least it makes a
				    -- few things go through that would
				    -- otherwise yield an error (e.g.
				    -- RemoveInv in the Queue example).
	    _		    -> Q.property False --error $ "Not a quickCheckable property: " ++ show goal
	where
	    toProp :: Value -> Q.Property
	    toProp = toHaskell

	    toGen :: Value -> Q.Gen Value
	    toGen = toHaskell

	    mkBoolProp v =
		case vAppView $ qcUnfoldConstants v of
		    VAVDef c _ []
			| c == qc_True  -> True
			| c == qc_False -> False
		    VAVDef c _ [_, v, w]
			| c == qc_Eq	-> v `equal` w
		    _			-> error $ "Not a decidable proposition: " ++ show v

instance Extractable (Q.Gen Value) where
    toHaskell v =
	case vAppProjView $ qcUnfoldConstants v of
	    VAPVDef c _ [IsProj VFst,_,IsApp w]
		| c == qc_monadGen  -> return (toHaskell w)
	    VAPVDef c _ [IsProj VSnd,_,_,IsApp gen,IsApp k]
		| c == qc_monadGen  -> do   v <- toHaskell gen
					    toHaskell (k `vApp` v)
	    VAPVDef c _ [_, IsApp w]
		| c == qc_elements  -> Q.elements (toHaskell w)
		| c == qc_oneof	    -> Q.oneof (toHaskell w)
		| c == qc_frequency -> Q.frequency $ toHaskell w
	    VAPVDef c _ [_, IsApp f]
		| c == qc_sized	    -> Q.sized (\n -> toHaskell $
						    f `vApp` fromHaskell v_Nat n
					    )
	    VAPVDef c _ [_, IsApp n, IsApp g]
		| c == qc_variant	-> Q.variant (toHaskell n) (toHaskell g)
-- TODO: primitive implementation of promote
{-	    VAPVDef c _ [IsApp a, IsApp b, IsApp f]
		| c == qc_promote	->
		    fromHaskell (VPi a (mkCl "_" a b) v_Set)
			<$> Q.promote (mkGenF f)
		where
		    mkGenF :: Value -> (Value -> Q.Gen Value)
		    mkGenF = toHaskell . toHaskell f
-}
	    _	-> error $ "Not a canonical generator: " ++ show v ++ "\n   " ++ show (qcUnfoldConstants v)
--	    _	-> Q.sized $ \n -> toHaskell v n <$> Q.rand

instance Extractable (StdGen -> Value) where
    toHaskell v r = genUnfoldConstants
			Recursive (\_ _ _ -> False)
			unfoldRand
			=<< vApp v v_seed
	where
	    unfoldRand v =
		case vAppView $ unfoldConstants v of
		    VAVDef c _ [_,xs,w]
			| c == qc_pick	-> Just $ pick (toHaskell xs) (toStdGen r w)
		    _			-> Nothing

-- Errors -----------------------------------------------------------------
-}
noInferenceForQCError :: ErrorFun
noInferenceForQCError e =
    e	{ errWhat   = fsep $ pwords "type inference failed"
	, errWhy    = fsep $ pwords "the type of a testable goal cannot be inferred"
	}

toManyArgumentsError :: Value -> ErrorFun
toManyArgumentsError v e =
    e	{ errWhat   = fsep $ pwords "QuickCheck was called with too many arguments"
	, errWhy    = fsep $ pwords "the goal" ++ [pretty v] ++
			     pwords "is not a function type"
	}

notAPropError :: Value -> ErrorFun
notAPropError v e =
    e	{ errWhat   = fsep $ pwords "Cannot quickCheck goal"
	, errWhy    = fsep $ pwords "the goal is" ++ [pretty v] ++
			     pwords "which is not a Prop"
	}

-- Help -------------------------------------------------------------------

helpQuickCheck = "The QuickCheck plugin.\n"
