{-| Defines the various error messages emitted by the type checker. -}
module TypeChecker.Errors where

import Prelude hiding	    ( print, putStr, putStrLn )
import IO.Output

import Text.PrettyPrint

import Syntax.Position
import Syntax.Internal
import Syntax.Pretty
import Syntax.Sorts
import Syntax.Values
import Utils.Error

-- Whats and whens --------------------------------------------------------

whenCheckingScope :: Doc
whenCheckingScope = fsep $ pwords "scope checking"

whenCheckingDefinition :: Name -> Doc
whenCheckingDefinition x =
    fsep $ pwords "checking the definition of" ++ [pretty x]

whenCheckingMutual :: Doc
whenCheckingMutual =
    fsep $ pwords $ "checking mutually recursive definitions"

whenCheckingType :: Term -> Type -> Doc
whenCheckingType e v =
    fsep $ pwords "checking that" ++ [pretty e] ++
	    pwords "has type" ++ [pretty v]

whenCheckingInheritance :: Name -> Name -> Doc
whenCheckingInheritance x y =
    fsep $ pwords "checking that" ++ [pretty x] ++
	   pwords "can inherit from" ++ [pretty y]

whenCheckingConstructor :: Name -> Doc
whenCheckingConstructor c =
    fsep $ pwords "checking the type of the constructor" ++ [pretty c]


whenInferringType :: Term -> Doc
whenInferringType e =
    fsep $ pwords "inferring the type of" ++ [pretty e]

whenTranslatingToFOL :: Value -> Doc
whenTranslatingToFOL v =
    fsep $ pwords "translating" ++ [pretty v] ++
	   pwords "to first order logic"

whenLoadingPlugin :: String -> Doc
whenLoadingPlugin name =
    fsep $ pwords "loading plugin" ++ [text name]

subSignatureMustHaveSameType :: Doc
subSignatureMustHaveSameType =
    fsep $ pwords "the types of the signatures are not the same"

checkingSubtypingConstraint :: Type -> Type -> Doc
checkingSubtypingConstraint w v =
    fsep $ pwords "the inferred type" ++ [pretty w] ++
	    pwords "is not a subtype of" ++ [pretty v]

doingNothingMuch :: Doc
doingNothingMuch = fsep $ pwords "you must have made a mistake"

notConstructorTypeOf :: Term -> Type -> Doc
notConstructorTypeOf v w =
    fsep $ pwords "a constructor of type" ++ [pretty v] ++
           pwords "cannot produce elements of data type" ++ [pretty w]

checkingPatternFailed :: Pattern -> Type -> Doc
checkingPatternFailed p v =
	fsep $ pwords "the pattern" ++ 
                      [pretty p] ++ 
                      pwords "does not have type" ++ [pretty v]

-- TODO: pretty print constructor application
checkingConstructorFailed :: Name -> [Term] -> Type -> Doc
checkingConstructorFailed c ts v =
    fsep $ pwords "constructed value" ++
           [pretty c] ++ map pretty ts ++ pwords "does not have type" ++
	   [pretty v]

-- Error messages ---------------------------------------------------------

customError :: String -> ErrorFun
customError s e = e { errWhy = fsep $ pwords s }

variablesNotEqualError :: Name -> Name -> ErrorFun
variablesNotEqualError x y e =
    e	{ errWhy = fsep $ [ pretty x ] ++ pwords "is not equal to" ++ [ pretty y ]
	}

variablesNotSubtypeError :: Name -> Name -> ErrorFun
variablesNotSubtypeError x y e =
    e	{ errWhy = fsep $ [ pretty x ] ++ pwords "is not a subtype of" ++ [ pretty y ]
	}

dontCareNotEqualError :: ErrorFun
dontCareNotEqualError e =
    e	{ errWhy = fsep $ pwords "_ is not equal to anything"
	}

sortsNotEqualError :: SortValue -> SortValue -> ErrorFun
sortsNotEqualError s1 s2 e =
    e	{ errWhy = fsep	[ pretty s1
			, text "is not equal to"
			, pretty s2
			]
	}

fieldsNotEqualError :: Name -> Name -> ErrorFun
fieldsNotEqualError x y e =
    e	{ errWhy = fsep	$ pwords "the field name" ++ [pretty x] ++
			  pwords "is not equal to" ++ [pretty y]
	}

sortsNotLeqError :: SortValue -> SortValue -> ErrorFun
sortsNotLeqError s1 s2 e =
    e	{ errWhy = fsep	[ pretty s1
			, text "is not a subsort of"
			, pretty s2
			]
	}

notAtAllEqualError :: Value -> Value -> ErrorFun
notAtAllEqualError u v e =
    e	{ errWhy = fsep	[ text "the value"
			, pretty u
			, text "is not equal to"
			, pretty v
			]
	}

variableNotBoundError :: Name -> ErrorFun
variableNotBoundError x e =
    e	{ errWhy = fsep $
		    pwords "the variable"
		    ++ [pretty x] ++
		    pwords "has not been bound"
	}

notAFunctionError :: Term -> Type -> ErrorFun
notAFunctionError e v m =
    m   { errWhy = fsep $
		    pwords "it has type"
		    ++ [pretty v] ++
		    pwords "which is not a function type"
	, errWhat = fsep $
		    pretty e :
		    pwords "is not a function"
	}

notAPairError :: Term -> Type -> ErrorFun
notAPairError e v m =
    m   { errWhy = fsep $
		    pwords "it has type"
		    ++ [pretty v] ++
		    pwords "which is not a sigma type"
	, errWhat = fsep $
		    pretty e :
		    pwords "is not a pair"
	}

notAStructError :: Term -> Value -> ErrorFun
notAStructError e v m =
    m   { errWhy = fsep $
		    pwords "it has type"
		    ++ [pretty v] ++
		    pwords "which is not a sigma type"
	, errWhat = fsep $
		    pretty e :
		    pwords "is not a pair"
	}

notATypeError :: Term -> Type -> ErrorFun
notATypeError e v m =
    m	{ errWhy = fsep $
		    pwords "it has type"
		    ++ [pretty v] ++
		    pwords "which is not a sort"
	, errWhat = fsep $
			pretty e :
			pwords "is not a type"
	}

notASortError :: Term -> ErrorFun
notASortError e m =
    m	{ errWhy = fsep $ pwords "I say so"
	, errWhat = fsep $ [pretty e] ++ pwords "is not a sort"
	}

noSuchFieldError :: String -> Value -> Name -> ErrorFun
noSuchFieldError x t s e =
    e	{ errWhat   = fsep $ pwords "could not project" ++ [text x] ++
			     pwords "from the record" ++ [pretty t]
	, errWhy    = fsep $ pwords "the signature" ++ [pretty s] ++
			     pwords "contains no such field"
	}

cannotInferError :: Term -> ErrorFun
cannotInferError e m =
    m	{ errWhat = fsep $ pwords "could not infer the type of"
			    ++ [pretty e]
	, errWhy = fsep $ pwords "it's not an inferrable thing (e.g. a lambda, constructor or pair)"
	}

lambdaHasFunctionTypeError :: Term -> Type -> ErrorFun
lambdaHasFunctionTypeError e v m =
    m	{ errWhat = fsep $ pretty e : pwords "does not have type" ++ [pretty v]
	, errWhy = fsep $ pwords "it has a function type"
	}

hasSigmaTypeError :: Term -> Type -> ErrorFun
hasSigmaTypeError e v m =
    m	{ errWhat   = fsep $ pretty e : pwords "does not have type" ++ [pretty v]
	, errWhy    = fsep $ pwords "it is a pair and thus its type is a sigma type"
	}

hasSignatureTypeError :: Term -> Type -> ErrorFun
hasSignatureTypeError e v m =
    m	{ errWhat   = fsep $ pretty e : pwords "does not have type" ++ [pretty v]
	, errWhy    = fsep $ pwords "it is a struct and thus its type is a signature"
	}

patternHasSigmaTypeError :: Pattern -> Type -> ErrorFun
patternHasSigmaTypeError e v m =
    m	{ errWhat   = fsep $ pretty e : pwords "does not have type" ++ [pretty v]
	, errWhy    = fsep $ pwords "it is a pair and thus its type is a sigma type"
	}

hasSigTypeError :: Term -> Type -> ErrorFun
hasSigTypeError e v m =
    m	{ errWhat   = fsep $ pretty e : pwords "does not have type" ++ [pretty v]
	, errWhy    = fsep $ pwords "it is a struct and thus its type is a signature"
	}

tooManyFieldsInStructError :: Name -> ErrorFun
tooManyFieldsInStructError x e =
    e	{ errWhat   = fsep $ pwords "too many fields in struct"
	, errWhy    = fsep $ pwords "the struct should end before the field"
			    ++ [pretty x]
	}

tooFewFieldsInStructError :: [Name] -> ErrorFun
tooFewFieldsInStructError xs e =
    e	{ errWhat   = fsep $ pwords "too few field in struct"
	, errWhy    = fsep $ pwords "the struct should have the fields"
			    ++ map pretty xs
	}

structLabelMismatchError :: Name -> Name -> ErrorFun
structLabelMismatchError x x' e =
    e	{ errWhat   = fsep $ pwords "field name mismatch"
	, errWhy    = fsep $ pwords "the field" ++ [pretty x] ++
			     pwords "was found where" ++ [pretty x'] ++
			     pwords "was expected"
	}

notDefinedInternalError :: Name -> ErrorFun
notDefinedInternalError x m =
    m	{ errWhat   = fsep $ pwords "there was an internal error"
	, errWhy    = fsep $ pwords "the definition of" ++ [pretty x] ++ pwords "is missing"
	}

repeatedVariableError :: Name -> ErrorFun
repeatedVariableError x e =
    e	{ errWhat   = fsep $ pwords "the name" ++ [pretty x] ++ pwords "is shadowed"
	}

noInferenceForPatternClausesError :: Clause -> ErrorFun
noInferenceForPatternClausesError c e =
    e	{ errWhat   = fsep $ pwords "cannot infer the type of the clause"
				++ [pretty c]
	, errWhy    = fsep $ pwords "it has a non-empty pattern list"
	}

clauseShouldHaveFunctionTypeError :: Clause -> Type -> ErrorFun
clauseShouldHaveFunctionTypeError c v e =
    e	{ errWhat   = fsep $ pwords "the clause" ++ [pretty c] ++
			     pwords "has too many arguments"
	, errWhy    = fsep $ [pretty v] ++ pwords "is not a function type"
	}

pluginNotFoundError :: String -> ErrorFun
pluginNotFoundError name e =
    e	{ errWhat   = fsep $ pwords "failed to call plugin"
	, errWhy    = fsep $ pwords "the plugin" ++ [text name] ++
			     pwords "does not exist"
	}

commandLineParseError :: String -> ErrorFun
commandLineParseError err e =
    e	{ errWhy    = text err
	, errCode   = OptionError
	}

emptySetError :: ErrorFun
emptySetError e =
    e	{ errWhat   = fsep $ pwords "sets must be non-empty"
	, errWhy    = fsep $ pwords "that's the way of the world"
	}

becauseWrongConstructor :: Name -> Name -> ErrorFun
becauseWrongConstructor c d e =
    e   { errWhy   = fsep $ [pretty c] ++ 
                            pwords "is not a constructor of data type" ++
                            [pretty d]
        }

becauseWrongNumberOfConstructorArguments :: Name -> ErrorFun
becauseWrongNumberOfConstructorArguments c e =
    e   { errWhy = fsep $ pwords "wrong number of arguments to constructor" 
                          ++ [pretty c]
        }

becauseNotADatatype :: Value -> ErrorFun
becauseNotADatatype v e =
    e   { errWhy    = fsep $ [pretty v] ++ pwords "is not a datatype"
        }

becauseNotASignature :: Value -> ErrorFun
becauseNotASignature v e =
    e   { errWhy    = fsep $ [pretty v] ++ pwords "is not a signature"
        }

cannotReconstructTopError :: ErrorFun
cannotReconstructTopError e =
    e	{ errWhat   = fsep $ pwords "term reconstruction failed"
	, errWhy    = fsep $ pwords "cannot reconstruct a term from the value Top"
	}

-- TODO: pretty print sort indices
sortInferenceUnificationError :: SortIndex -> SortIndex -> ErrorFun
sortInferenceUnificationError i1 i2 e =
    e	{ errWhat   = fsep $ pwords "inconsistency detected"
	, errWhy    = fsep $ [pretty i1] ++ pwords "is not equal to" ++ [pretty i2]
	}

unsolvedMetaError :: Term -> Int -> ErrorFun
unsolvedMetaError x n e =
    e	{ errWhat   = fsep $ pwords "failed to solve" ++ [text $ counter n] ++
			     pwords "hidden argument of" ++ [pretty x]
	, errWhy    = fsep $ pwords "nothing could be deduced about it"
	}

partiallySolvedMetaError :: Term -> Int -> Value -> ErrorFun
partiallySolvedMetaError x n v e =
    e	{ errWhat   = fsep $ pwords "failed to solve" ++ [text $ counter n] ++
			     pwords "hidden argument of" ++ [pretty x]
	, errWhy    = fsep $ pwords "the instantiation" ++ [pretty v] ++
			     pwords "is not complete"
	}

unsolvedUserMetaError :: Range -> ErrorFun
unsolvedUserMetaError r e =
    e	{ errWhat   = fsep $ pwords "failed to solve meta variable at" ++ [text $ show r]
	, errWhy    = fsep $ pwords "nothing could be deduced about it"
	}

partiallySolvedUserMetaError :: Range -> Value -> ErrorFun
partiallySolvedUserMetaError r v e =
    e	{ errWhat   = fsep $ pwords "failed to solve meta variable at" ++ [text $ show r]
	, errWhy    = fsep $ pwords "the instantiation" ++ [pretty v] ++
			     pwords "is not complete"
	}

hiddenApplicationError :: Term -> ErrorFun
hiddenApplicationError t e =
    e	{ errWhat   = fsep $ pwords "bad explicit application"
	, errWhy    = fsep $ [pretty t] ++ pwords "does not take implicit arguments"
	}

tooManyHiddenArguments :: Name -> Int -> Int -> ErrorFun
tooManyHiddenArguments x n m e =
    e	{ errWhat   = fsep $ pwords "too many hidden arguments for" ++ [pretty x]
	, errWhy    = fsep $ [pretty x] ++ pwords "takes only" ++ [pretty m] ++
			     pwords ("(not " ++ show n ++ ") hidden arguments")
	}

counter :: Int -> String
counter n = show n ++ suffixes !! n
    where
	suffixes = ["th","st","nd","rd"] ++ repeat "th"

